/* wptGPG.h - GnuPG interface
 *	Copyright (C) 2000-2005 Timo Schulz
 *
 * This file is part of WinPT.
 *
 * WinPT is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation; either version 2 
 * of the License, or (at your option) any later version.
 *  
 * WinPT is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License 
 * along with WinPT; if not, write to the Free Software Foundation, 
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 */

#ifndef WPT_GPG_H
#define WPT_GPG_H

#include "w32gpgme.h"
#include "wptKeyCache.h"
#include "wptNLS.h"


#define NO_STRICT 0

#define key_is_useable(key) (!(key)->revoked && !(key)->expired && !(key)->disabled)

enum setup_t {
    SETUP_KEYGEN    = 1,
    SETUP_IMPORT    = 2,
    SETUP_EXISTING  = 3
};

enum gpg_cmd_t {
    GPG_CMD_DECRYPT = 0,
    GPG_CMD_SIGN    = 1
};

enum keycache_t {
    KEYCACHE_PRV = 0,
    KEYCACHE_PUB = 1
};

enum entry_t {
    ENTRY_OPAQUE = 1,
    ENTRY_SINGLE = 2,
    ENTRY_MULTI  = 3,
    ENTRY_GROUP  = 4,
};

struct gpg_card_s;
typedef struct gpg_card_s * gpg_card_t;

/* GPG file association context. */
struct gpg_filetype {    
    const char *descr;
    const char *ext;
    int nicon;
};

/* Keycache refresh context. */
struct refresh_cache_s {
    int kr_reload;
    int kr_update;
    int tr_update;
};

struct gpg_option_s {
    struct gpg_option_s * next;
    char * name;
    char * val;
    unsigned int used:1;
    int type;  
};
typedef struct gpg_option_s * gpg_option_t;

struct gpg_member_s {
    struct gpg_member_s * next;
    char * name;
    unsigned int used:1;  
};
typedef struct gpg_member_s * gpg_member_t;

struct gpg_group_s {
    struct gpg_group_s * next;
    struct gpg_member_s * list;
    char * name;
    unsigned int used:1;  
};
typedef struct gpg_group_s * gpg_group_t;

struct gpg_optfile_s {
    struct gpg_option_s * list;
    struct gpg_group_s * grp;
};
typedef struct gpg_optfile_s * gpg_optfile_t;

struct passphrase_cb_s {
    int gpg_cmd;	/* sign or decrypt. */
    int pwd_init;	/* 1 = passphrase request via dialog. */
    char *pwd;		/* actual passphrase. */
    char info[1024];	/* hold up the info message */
    char *title;	/* dialog title. */
    char keyid[16+1];	/* keyid of the secret key the pwd can be used for. */
    HWND hwnd;		/* handle of the dialog window. */
    int cancel;		/* 1 = user cancelled passphrase request. */
    unsigned int is_card:1; /* 1 = key stored on a smart card. */
    unsigned int bad_pwd:1; /* 1 = last passphrase was bad. */
    gpgme_ctx_t gpg;
    gpgme_recipient_t recipients;
    struct {
	int sym_algo;
	int s2k_mode;
	int s2k_hash;
    } sym;
};


static const char * gpg_sigstat[] = {
    _("Error during verification process."),
    _("The signature is good."),
    _("The signature is BAD!"),
    _("The signature could not be checked due to a missing key."),
    _("No valid OpenPGP signature."),
    _("Signature Error"),
    _("Good Signature (Expired Key)"),
    _("Good Signature (Revoked Key)"),
    NULL
};

int sigsum_to_index (gpgme_sigsum_t sum);
#define SIGSTAT_MASK 8
#define get_gpg_sigstat(sum) gpg_sigstat[sigsum_to_index ((sum)) % SIGSTAT_MASK]

/* This variable is 1 if IDEA is available. */
extern int idea_available;

/*-- wptGPG.cpp --*/
char* get_gnupg_default_key( void );
int set_gnupg_default_key( const char *key );
char* get_gnupg_config (void);
char* get_gnupg_keyring_from_options( const char *fname, int pub );
int check_gnupg_config (const char *fname, int *secrings, int *pubrings);
char* get_gnupg_path( void );
int check_gnupg_prog( void );
int gnupg_access_files (void);
char* get_gnupg_prog( void );
char* get_gnupg_keyring (int pub, int strict);
char* get_gnupg_cfgfile (void);
int set_gnupg_options( const char *buf, size_t buflen );
int gnupg_load_config (void);

int check_gnupg_options (const char *buf);
void init_gnupg_table (void);
void free_gnupg_table (void);
int keyring_check_last_access (void);
const char* gnupg_check_file_ext (const char *fname, int *r_type);
int gpg_check_permissions (int showmsg);
int gnupg_check_homedir (void);
int gnupg_access_keyring (int _pub);
void gnupg_backup_options ();
void gnupg_backup_keyrings (void);
void gnupg_display_error (void);
int gnupg_copy_keyrings (void);
int check_gnupg_engine (int *r_major, int *r_minor, int *r_patch);

/*-- wptGPGME.cpp --*/
const char * get_signature_status( gpgme_sigsum_t sigstat );
void keycache_reload( HWND dlg );

int get_pubkey (const char *keyid, gpgme_key_t *ret_key);
int winpt_get_pubkey (const char *keyid, struct winpt_key_s *k);
int get_seckey (const char *keyid, gpgme_key_t *ret_skey);
int winpt_get_seckey (const char *keyid, struct winpt_key_s *k);

void keycache_reload( HWND dlg );
void keycache_release (int cleanup);
gpgme_error_t keycache_update (int is_sec, const char *keyid);
gpgme_error_t keycache_init( const char *pubring, const char * secring );
void keycache_set_reload( int yes );
int keycache_get_reload( void );
gpg_keycache_t keycache_get_ctx( int _pub );

int count_insecure_elgkeys (void);

int gpg_encrypt_symmetric (void);


/*-- wptGPGParser.cpp --*/
/* find */
gpg_group_t find_group( gpg_optfile_t opt, const char *str );
gpg_option_t find_option( gpg_optfile_t opt, const char *str );
gpg_member_t find_member( gpg_optfile_t opt, const char *grp, const char *str );
/* delete */
int delete_group( gpg_optfile_t opt, const char *str );
int delete_member( gpg_optfile_t opt, const char *grp, const char *str );
int delete_option( gpg_optfile_t opt, const char *str );
/* add */
int modify_entry( gpg_optfile_t opt, int type, const char *name, const char *val );
int add_entry( gpg_optfile_t opt, int type, const char *name, const char *val );
int add_member( gpg_optfile_t opt, const char *grp, const char *str );
int add_group( gpg_optfile_t opt, const char *str );
/* high-level */
int parse_gpg_options( const char *file, gpg_optfile_t *r_opt );
int commit_gpg_options( const char *file, gpg_optfile_t opt );
void release_gpg_options( gpg_optfile_t opt );
/* memory */
void release_group( gpg_group_t grp );

/*-- wptPassphraseCB.cpp --*/
enum passdlg_t {
    PASSDLG_REPEAT = 0,
    PASSDLG_INIT = 1,  
    PASSDLG_STRICT = 2
};

void set_gpg_passphrase_cb (passphrase_cb_s *cb, gpgme_ctx_t ctx,
			    int cmd, HWND hwnd, const char *title);
void release_gpg_passphrase_cb (passphrase_cb_s *cb);

char * request_passphrase (const char *title, int flags, int *ret_cancel);
char * request_passphrase2 (const char *title, int flags, int *ret_cancel);
const char * passphrase_cb (void *opaque, const char * desc, void *r_hd);
char * get_key_userid (const char *keyid);
int check_passwd_quality (const char *pass, int strict);

/*-- wptGPGUtil.cpp --*/
gpgme_error_t gpg_rebuild_cache (char **r_inf);
gpgme_error_t gpg_get_version (char **r_inf);
void          gpg_set_debug_mode (int val);
gpgme_error_t gpg_export_seckey (const char *keyid, const char *outfile);
gpgme_error_t gpg_manage_ownertrust (char **data, int do_export);

char*  generate_revoc_input (int code, const char *cmt, const char *pass);
gpgme_error_t gpg_revoke_key (const char *inp_data, 
			      const char *keyid,
			      char **r_revcert);
gpgme_error_t gpg_decode_c_string (const char *src, char **destp, size_t len);
gpgme_error_t gpg_import_key_list (const char *fname, char **r_out);
gpgme_error_t get_uat_validity (const char *keyid, gpgme_validity_t *r_valid);

#endif /* WPT_GPG_H */
