/* wptVerifyList.cpp - Listview for verifying signatures
 *	Copyright (C) 2001, 2002, 2003, 2005 Timo Schulz
 *	Copyright (C) 2005 g10 Code GmbH
 *
 * This file is part of WinPT.
 *
 * WinPT is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation; either version 2 
 * of the License, or (at your option) any later version.
 *  
 * WinPT is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License 
 * along with WinPT; if not, write to the Free Software Foundation, 
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <windows.h>
#include <time.h>

#include "wptTypes.h"
#include "wptGPG.h"
#include "wptCommonCtl.h"
#include "wptKeylist.h"
#include "wptNLS.h"
#include "wptErrors.h"
#include "wptUTF8.h"
#include "wptW32API.h"


/* Extract the file name part out of the given path in @path.
   Return value: file part or NULL on error. */
static char*
extract_filename (const char *path)
{
    char * fname, *p;
    int n, len = 0;

    p = strrchr (path, '\\');
    if (!p)
	return m_strdup (path);
    n = p - path;
    len = strlen (path) - n;
    fname = new char[len+1];
    if (!fname)
	BUG (NULL);
    memcpy (fname, path+n+1, len);
    fname[len] = '\0';
    return fname;
}


/* String representaton of the time in @timestamp.
   Format YEAR-MON-DAY HOUR:MIN:SEC.
   Return value: time as formatted string. */
const char *
strtimestamp (long timestamp)
{
    static char timebuf[64] = {0};
    struct tm *warp;
    
    warp = localtime (&timestamp);
    _snprintf (timebuf, sizeof timebuf - 1, "%04d-%02d-%02d %02d:%02d:%02d",
               warp->tm_year+1900, warp->tm_mon+1, warp->tm_mday,	
               warp->tm_hour, warp->tm_min, warp->tm_sec);
    return timebuf;
}


/* Map the signature summary in @sum to signature status table index. 
   Return value: index to table. */
int
sigsum_to_index (gpgme_sigsum_t sum)
{
    if ((sum & GPGME_SIGSUM_VALID) && (sum & GPGME_SIGSUM_KEY_REVOKED))
	return 7;
    if ((sum & GPGME_SIGSUM_VALID) && (sum & GPGME_SIGSUM_SIG_EXPIRED))
	return 6;
    if (sum & GPGME_SIGSUM_GREEN)
	return 1;
    else if (sum & GPGME_SIGSUM_RED)
	return 2;
    else if (sum & GPGME_SIGSUM_KEY_MISSING)
	return 3;
    return 0;
}


/* Build a verify signature list control. With the parent window
   from @ctrl and the mod given in @fm_mode. @lv contains the
   new control on success.
   Return value: 0 on success. */
int
verlist_build (listview_ctrl_t *lv, HWND ctrl, int fm_mode)
{
    int j, rc = 0;
    struct listview_ctrl_s * c;
    struct listview_column_s fm_verlist[] = {	
	{0, 100, (char *)_("Name") },
	{1, 120, (char *)_("Status") },
	{2, 115, (char *)_("Signed") },
	{3,  58, (char *)_("Trust") },		
	{4, 160, (char *)_("User ID") },
	{5, 0, NULL}
	
    };
    struct listview_column_s verlist[] = {	
	{0, 140, (char *)_("Status") },
	{1, 120, (char *)_("Signed") },
	{2,  58, (char *)_("Trust") },
	{3,  80, (char *)_("Key ID" )},
	{4, 160, (char *)_("User ID") },
	{5, 0, NULL}
    };
    
    rc = listview_new (&c);
    if (rc)
	return rc;
    
    c->ctrl = ctrl;
    if( fm_mode ) {
	for( j = 0; fm_verlist[j].fieldname; j++ )	
	    listview_add_column( c, &fm_verlist[j] );
    }
    else {
	for( j = 0; verlist[j].fieldname; j++ )
	    listview_add_column( c, &verlist[ j ] );	
    }
    listview_set_ext_style( c );
    *lv = c;
    return 0;
}


/* Delete the given verify control in @lv. */
void
verlist_delete (listview_ctrl_t lv)
{
    if (lv) {
	listview_release (lv);	
    }
}


/* Add the given signature in @sig to the verify control @lv.
   Return value: 0 on success. */
int
verlist_add_sig (listview_ctrl_t lv, gpgme_signature_t sig)
{
    gpgme_key_t key = NULL;
    const char * attr;
    u32 key_attr;
    char keyid[32+1];
    char * uid = NULL;
    
    if (listview_add_item (lv, " "))
	return WPTERR_GENERAL;
    
    get_pubkey (sig->fpr, &key);

    if (sig->summary == 0 && gpg_err_code (sig->status) == GPG_ERR_NO_ERROR)
	attr = get_gpg_sigstat (GPGME_SIGSUM_GREEN);
    else
	attr = get_gpg_sigstat (sig->summary);
    if (attr)
	listview_add_sub_item (lv, 0, 0, (char *)attr);
    
    attr = strtimestamp (sig->timestamp);
    listview_add_sub_item (lv, 0, 1, (char *)attr);
    
    attr = _("Unknown");
    if (key) {
	key_attr = key->uids->validity;
	attr = get_key_trust2 (NULL, key_attr, 0, 0);
    }
    listview_add_sub_item (lv, 0, 2, (char *)attr);
    
    attr = sig->fpr;
    if (!attr || strlen (attr) < 8)
	listview_add_sub_item (lv, 0, 3, "????????");
    else {
	if (strlen (attr) == 40)
	    attr += 32;
	else if (strlen (attr) == 32)
	    attr += 24;
	else
	    attr += 8;
	_snprintf (keyid, sizeof keyid -1, "0x%s", attr);
	listview_add_sub_item (lv, 0, 3, keyid);
    }
    
    if (!key) {
	attr = _("Invalid User ID");
	listview_add_sub_item( lv, 0, 4, (char *)attr );
    }
    else {
	attr = key->uids->name;
	uid = utf8_to_wincp (attr, strlen (attr));
	if (uid) {
	    listview_add_sub_item( lv, 0, 4, (char *)uid );
	    free (uid);
	}
    }

    return 0;
}


/* Add the given file signature in @log to the verify control @lv.
   Return value: 0 on success. */
int
verlist_add_sig_log (listview_ctrl_t lv, file_sig_ctx_t log)
{
    gpgme_signature_t sig = log->sig;
    gpgme_key_t key = NULL;
    const char *attr;    
    char t[64], *name;

    if (listview_add_item (lv, ""))
	return WPTERR_GENERAL;

    get_pubkey (sig->fpr, &key);
    
    name = extract_filename (log->file);
    if (name)
	listview_add_sub_item (lv, 0, 0, name);
    else
	listview_add_sub_item (lv, 0, 0, log->file);
    free_if_alloc (name);

    if (sig->summary == 0 && gpg_err_code (sig->status) == GPG_ERR_NO_ERROR)
	attr = get_gpg_sigstat (GPGME_SIGSUM_GREEN);
    else
	attr = get_gpg_sigstat (sig->summary);
    if (attr)
	listview_add_sub_item (lv, 0, 1, attr);

    if (sig->timestamp > 0) {
	attr = strtimestamp (sig->timestamp);
	listview_add_sub_item (lv, 0, 2, attr);
    }
    else
	listview_add_sub_item (lv, 0, 2, "No time");
    
    attr = _("Unknown");
    if (key)
	attr = get_key_trust2 (NULL, key->uids->validity, 0, 0);
    listview_add_sub_item (lv, 0, 3, attr);
    
    attr = sig->fpr;
    if (!log->use_uid && strlen (attr) == 40) {
	_snprintf (t, sizeof (t)-1, "0x%s", attr + 32);
	listview_add_sub_item (lv, 0, 4, t);
    }
    else if( !log->use_uid && strlen( attr ) == 32 ) {
	_snprintf (t, sizeof (t)-1, "0x%s", attr + 24);
	listview_add_sub_item (lv, 0, 4, t);
    }
    else if (log->user_id) {
	char *p = utf8_to_wincp (log->user_id, strlen (log->user_id));
	if (p) {
	    listview_add_sub_item (lv, 0, 4, p);
	    free (p);
	}
    }
    return 0;
}
